@@ -2,6 +2,7 @@ import os
 import re
 from abc import ABC, abstractmethod
 import random
+import time
 
 import openai
 
@@ -58,12 +59,26 @@ class DummyManuscriptRevisionModel(ManuscriptRevisionModel):
                 + "\n%%% PARAGRAPH END %%%"
             )
 
-        return self.sentence_end_pattern.sub(" ", paragraph_text).strip()
+        return self.sentence_end_pattern.sub(". ", paragraph_text).strip()
 
     def get_prompt(self, paragraph_text, section_name):
         return paragraph_text
 
 
+class VerboseManuscriptRevisionModel(DummyManuscriptRevisionModel):
+    """
+    This model returns the same paragraph and adds a header to it.
+    """
+
+    def __init__(self, revised_header: str = "Revised:"):
+        super().__init__()
+        self.revised_header = revised_header
+
+    def revise_paragraph(self, paragraph_text, section_name):
+        revised_paragraph = super().revise_paragraph(paragraph_text, section_name)
+        return f"{self.revised_header}{revised_paragraph}"
+
+
 class RandomManuscriptRevisionModel(ManuscriptRevisionModel):
     """
     This model takes a paragraph and randomizes the words. The paragraph has the
@@ -117,7 +132,8 @@ class GPT3CompletionModel(ManuscriptRevisionModel):
         frequency_penalty: float = None,
         best_of: int = None,
         top_p: float = None,
-        retry_count: int = 3,
+        retry_count: int = 5,
+        edit_endpoint: bool = False,
     ):
         super().__init__()
 
@@ -200,6 +216,16 @@ class GPT3CompletionModel(ManuscriptRevisionModel):
         self.title = title
         self.keywords = keywords
 
+        # adjust options if edits endpoint was selected
+        self.edit_endpoint = edit_endpoint
+        if self.edit_endpoint and model_engine == "text-davinci-003":
+            model_engine = "text-davinci-edit-001"
+
+        if model_engine == "text-davinci-edit-001":
+            self.edit_endpoint = True
+
+        print("Language model: ", model_engine)
+
         self.model_parameters = {
             "engine": model_engine,
             "temperature": temperature,
@@ -218,56 +244,91 @@ class GPT3CompletionModel(ManuscriptRevisionModel):
 
         self.several_spaces_pattern = re.compile(r"\s+")
 
-    def get_prompt(self, paragraph_text: str, section_name: str) -> str:
+    def get_prompt(
+        self, paragraph_text: str, section_name: str
+    ) -> str | tuple[str, str]:
         """
-        Returns the prompt to be used for the revision of a paragraph that belongs
-        to a given section. There are three types of prompts according to the section:
-        Abstract, Introduction and the rest (i.e. Methods, Results, Discussion, etc.).
+        Returns the prompt to be used for the revision of a paragraph that
+        belongs to a given section. There are three types of prompts according
+        to the section: Abstract, Introduction, Methods, and the rest (i.e.
+        Results, Discussion, etc.).
 
         Args:
             paragraph_text: text of the paragraph to revise.
             section_name: name of the section the paragraph belongs to.
 
         Returns:
-            Prompt to be used by the model for the revision of the paragraph.
+            If self.edit_endpoint is False, then returns a string with the prompt to be used by the model for the revision of the paragraph.
             It contains two paragraphs of text: the command for the model
             ("Revise...") and the paragraph to revise.
+
+            If self.edit_endpoint is True, then returns a tuple with two strings:
+             1) the instructions to be used by the model for the revision of the paragraph,
+             2) the paragraph to revise.
         """
         if section_name in ("abstract",):
             prompt = f"""
-                Revise the following paragraph from the {section_name} of an academic paper with the title '{self.title}' and keywords '{", ".join(self.keywords)}.'
-                Make sure the paragraph is easy to read, it is in active voice, and the take-home message is clear:
+                Revise the following paragraph from the {section_name} of an academic paper (with the title '{self.title}' and keywords '{", ".join(self.keywords)}')
+                so the research problem/question is clear,
+                   the solution proposed is clear,
+                   the text grammar is correct, spelling errors are fixed,
+                   and the text is in active voice and has a clear sentence structure
             """
-        elif section_name in ("introduction",):
+        elif section_name in ("introduction", "discussion"):
             prompt = f"""
-                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper with the title '{self.title}' and keywords '{", ".join(self.keywords)}.'
-                Make sure the paragraph has a clear and easy-to-read sentence structure, it is in active voice, and it minimizes the use of jargon.
-                Keep most of the citations to other academic papers:
+                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper (with the title '{self.title}' and keywords '{", ".join(self.keywords)}')
+                so
+                   most of the citations to other academic papers are kept,
+                   the text minimizes the use of jargon,
+                   the text grammar is correct, spelling errors are fixed,
+                   and the text has a clear sentence structure
+            """
+        elif section_name in ("results",):
+            prompt = f"""
+                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper (with the title '{self.title}' and keywords '{", ".join(self.keywords)}')
+                so
+                   most references to figures and tables are kept,
+                   the details are enough to clearly explain the outcomes,
+                   sentences are concise and to the point,
+                   the text minimizes the use of jargon,
+                   the text grammar is correct, spelling errors are fixed,
+                   and the text has a clear sentence structure
             """
         elif section_name in ("methods",):
-            prompt = f"""
-                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper with the title '{self.title}' and keywords '{", ".join(self.keywords)}.'
-                Preserve the math. Keep all the equations
+            equation_definition = r"$$ ... $$ {#id}"
+            revise_sentence = f"""
+                Revise the paragraph(s) below from
+                the {section_name.capitalize()} section of an academic paper
+                (with the title '{self.title}' and keywords '{", ".join(self.keywords)}')
             """.strip()
 
-            if "$$" not in paragraph_text:
-                prompt = f"""
-                    {prompt}.
-                    Make sure the paragraph has a clear and easy-to-read sentence structure, and it minimizes the use of jargon:
-                """
-            else:
-                prompt = f"""
-                    {prompt}:
-                """
+            prompt = f"""
+                {revise_sentence}
+                so
+                   most of the citations to other academic papers are kept,
+                   most of the technical details are kept,
+                   most references to equations (such as "Equation (@id)") are kept,
+                   all equations definitions (such as '{equation_definition}') are included with newlines before and after,
+                   the most important symbols in equations are defined,
+                   spelling errors are fixed, the text grammar is correct,
+                   and the text has a clear sentence structure
+            """.strip()
         else:
             prompt = f"""
-                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper with the title '{self.title}' and keywords '{", ".join(self.keywords)}.'
-                Make sure the paragraph has a clear and easy-to-read sentence structure, and it minimizes the use of jargon:
+                Revise the following paragraph from the {section_name.capitalize()} section of an academic paper (with the title '{self.title}' and keywords '{", ".join(self.keywords)}')
+                so
+                   the text minimizes the use of jargon,
+                   the text grammar is correct, spelling errors are fixed,
+                   and the text has a clear sentence structure
             """
 
         prompt = self.several_spaces_pattern.sub(" ", prompt).strip()
 
-        return f"{prompt}\n\n{paragraph_text.strip()}"
+        if not self.edit_endpoint:
+            return f"{prompt}.\n\n{paragraph_text.strip()}"
+        else:
+            prompt = prompt.replace("the following paragraph", "this paragraph")
+            return f"{prompt}.", paragraph_text.strip()
 
     def get_max_tokens(self, paragraph_text: str, fraction: float = 2.0) -> int:
         """
@@ -357,34 +418,63 @@ class GPT3CompletionModel(ManuscriptRevisionModel):
         prompt = self.get_prompt(paragraph_text, section_name)
 
         params = {
-            "prompt": prompt,
-            "max_tokens": max_tokens,
-            "stop": None,
             "n": 1,
         }
 
+        if self.edit_endpoint:
+            params.update(
+                {
+                    "instruction": prompt[0],
+                    "input": prompt[1],
+                }
+            )
+        else:
+            params.update(
+                {
+                    "prompt": prompt,
+                    "max_tokens": max_tokens,
+                    "stop": None,
+                }
+            )
+
         params.update(self.model_parameters)
 
         retry_count = 0
         message = ""
         while message == "" and retry_count < self.retry_count:
             try:
-                completions = openai.Completion.create(**params)
+                print(
+                    f"[Attempt #{retry_count}] Revising paragraph '{paragraph_text[:20]}'..."
+                )
+
+                if self.edit_endpoint:
+                    completions = openai.Edit.create(**params)
+                else:
+                    completions = openai.Completion.create(**params)
+
                 message = completions.choices[0].text.strip()
             except Exception as e:
                 error_message = str(e)
-                token_stats = self.get_max_tokens_from_error_message(error_message)
+                print(f"Error: {error_message}")
 
-                if token_stats is None:
-                    raise e
+                # if the error message suggests to sample again, let's do that
+                if "Please sample again" in error_message:
+                    pass
+                elif "overloaded" in error_message:
+                    time.sleep(5)
+                else:
+                    # if the error mesaage suggests to reduce the number of tokens,
+                    # obtain the number of tokens to reduce and retry
+                    token_stats = self.get_max_tokens_from_error_message(error_message)
 
-                max_context_length = token_stats["max_context_length"]
-                # requested_tokens = token_stats["requested_tokens"]
-                tokens_in_prompt = token_stats["tokens_in_prompt"]
-                # tokens_in_completion = token_stats["tokens_in_completion"]
+                    if token_stats is None:
+                        raise e
 
-                params["max_tokens"] = max_context_length - tokens_in_prompt
+                    max_context_length = token_stats["max_context_length"]
+                    tokens_in_prompt = token_stats["tokens_in_prompt"]
 
-            retry_count += 1
+                    params["max_tokens"] = max_context_length - tokens_in_prompt
+            finally:
+                retry_count += 1
 
         return message
